# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Coral face recognition in image poll plugin """

import copy
import json
import uuid
import logging
import time
import re
import os
from PIL import Image

from foglamp.common import logger
from foglamp.plugins.common import utils
from edgetpu.detection.engine import DetectionEngine

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2019 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Plugin to detect faces in image',
        'type': 'string',
        'default': 'coral_people_image',
        'readonly': 'true'
    },
    'model_file': {
        'description': 'TFlite model file to use for inference',
        'type': 'string',
        'default': 'mobilenet_ssd_v2_face_quant_postprocess_edgetpu.tflite',
        'order': '1',
        'displayName': 'TFlite model file'
    },
    'labels_file': {
        'description': 'Labels file used during inference',
        'type': 'string',
        'default': 'coco_labels.txt',
        'order': '2',
        'displayName': 'Labels file'
    },
    'assetName': {
        'description': 'Asset Name',
        'type': 'string',
        'default': 'CoralFaceDetection',
        'order': '3',
        'displayName': 'Asset Name'
    },
    'image_file': {
        'description': 'Image file to classify',
        'type': 'string',
        'default': '/tmp/cap.jpg',
        'order': '4',
        'displayName': 'Image file path'
    }
}
_LOGGER = logger.setup(__name__, level=logging.INFO)


engine = None
labels = None
asset_name = None
threshold = None
top_k = None


def load_labels(path):
    p = re.compile(r'\s*(\d+)(.+)')
    with open(path, 'r', encoding='utf-8') as f:
        lines = (p.match(line).groups() for line in f.readlines())
        return {int(num): text.strip() for num, text in lines}


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """

    return {
        'name': 'Face recognition image plugin',
        'version': '1.7.0',
        'mode': 'poll',
        'type': 'south',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Initialise the plugin with the merged config
        Args:
           config: JSON configuration document for the South plugin configuration category
        Returns:
           data: JSON object to be used in future calls to the plugin
        Raises:
    """
    global engine, labels, asset_name, threshold, top_k
    _LOGGER.debug("plugin_init: config = %s", json.dumps(config))
    try:
        model = config['model_file']['value']
        labels = config['labels_file']['value']
        asset_name = config['assetName']['value']
        threshold = 0.1
        top_k = 10

        model = os.path.join(os.path.dirname(__file__), "model", model)
        labels = os.path.join(os.path.dirname(__file__), "model", labels)

        _LOGGER.debug("Loading {} with {} labels.".format(model, labels))
        engine = DetectionEngine(model)
        labels = load_labels(labels)

    except Exception as e:
        _LOGGER.exception("Coral face recognition in image plugin failed to start. Details %s", str(e))
        raise

    data = copy.deepcopy(config)
    return data


def plugin_poll(handle):
    """ Run inference on given image and returns a reading object in a JSON document.
    Available for poll mode only.
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        returns a inference based reading in a JSON document, if it is available
        None - If no reading is available
    Raises:
        Exception
    """
    global engine, labels, asset_name, threshold, top_k
    try:
        image = Image.open(handle['image_file']['value'])
        image.show()

        start_time = time.monotonic()
        # DetectWithImage will resize the image as per input tensor shape
        objs = engine.DetectWithImage(image, threshold=threshold,
                                      keep_aspect_ratio=True, relative_coord=True,
                                      top_k=top_k)

        end_time = time.monotonic()
        _LOGGER.debug('Inference time: {:0.2f} ms'.format((end_time - start_time) * 1000))

        coords = []
        width, height = image.size
        for obj in objs:
            x0, y0, x1, y1 = obj.bounding_box.flatten().tolist()
            x0, y0, x1, y1 = int(x0 * width), int(y0 * height), int(x1 * width), int(y1 * height)
            score = int(100 * obj.score)
            coords.append({'corners': [[x0, y0], [x1, y1]], 'score': score})

        reads = {
            'count': len(objs),
            'coords': coords
        }
        data = {
            'asset': asset_name,
            'timestamp': utils.local_timestamp(),
            'key': str(uuid.uuid4()),
            'readings': reads
        }

        if len(objs) > 0:
            _LOGGER.debug(json.dumps(data))
            return data

    except (Exception, RuntimeError, FileNotFoundError) as ex:
        _LOGGER.exception("Coral face recognition in image plugin exception: {}".format(str(ex)))
        return None
    else:
        return data


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    it should be called when the configuration of the plugin is changed during the operation of the south service.
    The new configuration category should be passed.

    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    Raises:
    """
    _LOGGER.debug("Old config for Coral face recognition in image plugin {} \n new config {}".format(handle, new_config))

    # TODO: limit complete plugin restart (shutdown + start) to only the cases where absolutely necessary
    plugin_shutdown(handle)
    new_handle = copy.deepcopy(new_config)
    plugin_init(new_handle)
    return new_handle


def plugin_shutdown(handle):
    global engine, labels, asset_name, threshold, top_k
    try:
        _LOGGER.info('Coral face recognition in image plugin shutting down.')
        engine = None
        labels = None
        asset_name = None
        threshold = None
        top_k = None
    except Exception as e:
        _LOGGER.exception(str(e))
        raise
